import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import torch.backends.cudnn as cudnn

import torchvision
import torchvision.transforms as transforms
import torchvision.transforms.functional as TF
import numpy as np

import os
import time
import math
from tqdm import tqdm
from math import ceil
import matplotlib.pyplot as plt
import matplotlib

from ._models import *
from ._utils_dre import *


# os.environ['MKL_DEBUG_CPU_TYPE'] = '5'  # 3900x


class DeepRecursiveEmbedding:
    """
    	****** Deep Recursive Embedding implementation (PyTorch). ******
		Author: Xinrui Zu
		Version: 1.0
		Date: 2021/10/10

		Original Paper:
		Authors: Zixia Zhou, Xinrui Zu, Yuanyuan Wang, Boudewijn P.F. Lelieveldt, Qian Tao

        n_components:
        The embedding dimensionality.

        n_pre_epochs:
        The number of epochs before the recursive steps.

        n_recursive_epochs:
        The number of epochs in each recursive step. (default total steps: 300+100*4)

        dre_type (default 'fc'):
        The default type of the network. options: ['fc', 'conv']

        learning_rate:
        The learning rate of the optimization.

        tsne_perplexity (default: 30.0):
        The parameter defininig the normalization of t-SNE Pij.

        umap_n_neighbors (default: 15):
        The number of neighbors when calculating Pij in the last UMAP-like recursive step.

        min_dist:
        The minimum distance between the embedding points. Used to form Qij.

        batch_size (default: 2500):
        The batch size of the training procedure.

        rebatching_epochs:
		The number of epochs when the minibatches are fixed in the total dataset.

		data_dim:
		The dimensionality of the input data (when using fully-connected DRE).

		data_dim_conv:
		The dimensionality of the input tensor (when using convolutional DRE).
		Example: Fashion-MNIST: [3,32,32]
    """
    def __init__(self,
                 n_components=2,
                 n_pre_epochs=300,
                 n_recursive_epochs=150,
                 dre_type='fc',  # 'fc' or 'conv'
                 learning_rate=5.0 * 1e-5,
                 tsne_perplexity=30.0,
                 umap_n_neighbors=15,
                 min_dist=0.001,
                 batch_size=2500,
                 rebatching_epochs=1e4,
                 data_dim=0,
                 data_dim_conv=[],
                 save_steps=False,
                 plot_results=False,
                 directory='~/',
                 ):
        self.dre_type = dre_type
        self.learn_from_exist = False

        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.lowest_loss = 10000.0
        self.num_pre_epochs = n_pre_epochs
        self.num_recursive_epochs = n_recursive_epochs
        self.num_epochs = self.num_pre_epochs + 4*self.num_recursive_epochs
        self.batch_size = batch_size

        self.perplexity = tsne_perplexity
        self.umap_knn = umap_n_neighbors
        self.min_dist = min_dist
        self.a, self.b = find_ab_params(1, self.min_dist)

        self.lr = learning_rate

        self.rolling_num = rebatching_epochs
        self.data_dim = data_dim
        self.data_dim_conv = data_dim_conv

        self.P = []
        self.data = 0
        self.num_batch = 0
        self.embedding = 0
        self.save_steps = save_steps
        self.plot_results = plot_results
        self.directory = directory

        self.loss_plot_train = []
        self.loss_plot_val = []

        # self.num_batch = int(self.data.shape[0] / self.batch_size)
        # self.n = self.num_batch * self.batch_size

        # self.start_time = time.time()
        if not os.path.isdir('checkpoint'):
        	os.mkdir('checkpoint')

        print('<================ Building model ================>')

        if self.dre_type == 'conv':
            self.net = DREConvSmall(*self.data_dim_conv)
        elif self.dre_type == 'fc':
            self.net = DRE(self.data_dim)
        else:
            raise TypeError('[DRE] DRE type must be conv or normal.')

        self.net = self.net.to(self.device)
        if self.device == 'cuda':
            self.net = torch.nn.DataParallel(self.net)
            cudnn.benchmark = True
        self.net_optim = 0

        print('<================ Model loaded ================>')

        # Loss function and optimization method:

        self.optimizer = optim.Adam(self.net.parameters(), lr=self.lr)  # 5.0*(1e-5) for the original code
        # self.optimizer = optim.SGD(net.parameters(), lr=1e-4,
        #                       momentum=0.9, weight_decay=5e-4)
        # self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(self.optimizer, T_max=1500)

    def calculate_p_matrix(self, step):
        # if self.data == 0:
        #     raise TypeError('[DRE] Input the data first.')
        ran_num = np.random.randint(2 ** 16 - 1)
        np.random.seed(ran_num)
        np.random.shuffle(self.data)
        return self._calculate_p_matrix(step)  # fill in 'pre', 're1', 're2', 're3', 're_umap'

    def _calculate_p_matrix(self, step):
        print('building P matrix...')
        self.P = []
        self.net.eval()
        with torch.no_grad():
            if step == 'pre':
                for i in range(self.num_batch):
                    inputs = self.data[i * self.batch_size:(i + 1) * self.batch_size].astype('float32')
                    P1 = conditional_probability_p(self.perplexity, self.umap_knn, inputs, p_type='tsne')
                    self.P.append(P1)
                    print('[DRE] P length: ', len(self.P))
            if step == 're1':
                for i in range(self.num_batch):
                    if self.dre_type == 'conv':
                        inputs = self.data[i * self.batch_size:(i + 1) * self.batch_size].astype('float32')
                        inputs = torch.from_numpy(inputs).permute(0, 3, 1, 2)
                    elif self.dre_type == 'fc':
                        inputs = self.data[i * self.batch_size:(i + 1) * self.batch_size].astype('float32')
                        inputs = torch.from_numpy(inputs)
                    else:
                        raise TypeError('DRE type must be conv or fc.')
                    low_dim_data, _, _, _ = self.net(inputs)
                    low_dim_data = np.array(low_dim_data.cpu())
                    print('[DRE] recursive step 1 data shape: ', low_dim_data.shape)
                    P1 = conditional_probability_p(self.perplexity, self.umap_knn, low_dim_data, p_type='tsne')
                    self.P.append(P1)
                    print('[DRE] P length: ', len(self.P))
            if step == 're2':
                for i in range(self.num_batch):
                    if self.dre_type == 'conv':
                        inputs = self.data[i * self.batch_size:(i + 1) * self.batch_size].astype('float32')
                        inputs = torch.from_numpy(inputs).permute(0, 3, 1, 2)
                    elif self.dre_type == 'fc':
                        inputs = self.data[i * self.batch_size:(i + 1) * self.batch_size].astype('float32')
                        inputs = torch.from_numpy(inputs)
                    else:
                        raise TypeError('DRE type must be conv or fc.')
                    _, low_dim_data, _, _ = self.net(inputs)
                    low_dim_data = np.array(low_dim_data.cpu())
                    print('[DRE] recursive step 2 data shape: ', low_dim_data.shape)
                    P1 = conditional_probability_p(self.perplexity, self.umap_knn, low_dim_data, p_type='tsne')
                    self.P.append(P1)
                    print('[DRE] P length: ', len(self.P))
            if step == 're3':
                for i in range(self.num_batch):
                    if self.dre_type == 'conv':
                        inputs = self.data[i * self.batch_size:(i + 1) * self.batch_size].astype('float32')
                        inputs = torch.from_numpy(inputs).permute(0, 3, 1, 2)
                    elif self.dre_type == 'fc':
                        inputs = self.data[i * self.batch_size:(i + 1) * self.batch_size].astype('float32')
                        inputs = torch.from_numpy(inputs)
                    else:
                        raise TypeError('DRE type must be conv or fc.')
                    _, _, low_dim_data, _ = self.net(inputs)
                    low_dim_data = np.array(low_dim_data.cpu())
                    print('[DRE] recursive step 3 data shape: ', low_dim_data.shape)
                    P1 = conditional_probability_p(self.perplexity, self.umap_knn, low_dim_data, p_type='tsne')
                    self.P.append(P1)
                    print('[DRE] P length: ', len(self.P))
            if step == 're_umap':
                for i in range(self.num_batch):
                    if self.dre_type == 'conv':
                        inputs = self.data[i * self.batch_size:(i + 1) * self.batch_size].astype('float32')
                        inputs = torch.from_numpy(inputs).permute(0, 3, 1, 2)
                    elif self.dre_type == 'fc':
                        inputs = self.data[i * self.batch_size:(i + 1) * self.batch_size].astype('float32')
                        inputs = torch.from_numpy(inputs)
                    else:
                        raise TypeError('DRE type must be conv or fc.')
                    _, _, low_dim_data, _ = self.net(inputs)
                    low_dim_data = np.array(low_dim_data.cpu())
                    print('[DRE] recursive step 3 data shape: ', low_dim_data.shape)
                    P1 = conditional_probability_p(self.perplexity, self.umap_knn, low_dim_data, p_type='umap')
                    self.P.append(P1)
                    print('[DRE] P length: ', len(self.P))

        return self.P

    def _train(self, epoch, step):
        print('\nEpoch: %d' % epoch)
        self.net.train()  # train mode
        loss_all = 0
        with tqdm(total=self.data.shape[0], desc=f'[DRE] Training.. Epoch {epoch + 1}/{self.num_epochs}', unit='img', colour='blue') as pbar:
            for i in range(self.num_batch):
                # Load the packed data:
                tar = self.P[i]
                inputs, targets = torch.from_numpy(self.data[i*self.batch_size:(i+1)*self.batch_size]).type(torch.FloatTensor).to(self.device),\
                                  torch.from_numpy(tar).type(torch.FloatTensor).to(self.device)  # 如果是long会全变成0

                inputs = inputs.reshape(inputs.shape[0], -1)
                self.optimizer.zero_grad()
                _, _, _, outputs = self.net(inputs)

                loss = loss_function(targets, outputs, self.a, self.b, type=step)
                # print(loss)
                if str(loss.item()) == str(np.nan):
                    print('[DRE] detect nan in loss function, skip this iter')
                    continue
                loss.backward()
                loss_all += loss.item()
                # losses[divide_number*i+j] = loss
                self.optimizer.step()
                pbar.set_postfix(**{'loss (batch)': loss.item()})
                pbar.update(targets.shape[0])
        return loss_all / self.num_batch

    def _train_conv(self, epoch, step):
        print('\nEpoch: %d' % epoch)
        self.net.train()  # train mode
        loss_all = 0
        with tqdm(total=self.data.shape[0], desc=f'[DRE] Training.. Epoch {epoch + 1}/{self.num_epochs}', unit='img', colour='blue') as pbar:
            for i in range(self.num_batch):
                # Load the packed data:
                tar = self.P[i]
                inputs, targets = torch.from_numpy(self.data[i*self.batch_size:(i+1)*self.batch_size])\
                                      .type(torch.FloatTensor).permute(0, 3, 1, 2).to(self.device),\
                                  torch.from_numpy(tar).type(torch.FloatTensor).to(self.device)  # 如果是long会全变成0
                self.optimizer.zero_grad()
                _, _, _, outputs = self.net(inputs)

                loss = loss_function(targets, outputs, self.a, self.b, type=step)
                if str(loss.item()) == str(np.nan):
                    print('[DRE] detect nan in loss function, skip this iter')
                    continue
                loss.backward()
                loss_all += loss.item()
                # losses[divide_number*i+j] = loss
                self.optimizer.step()
                pbar.set_postfix(**{'loss (batch)': loss.item()})
                pbar.update(targets.shape[0])
        return loss_all / self.num_batch

    def _validation(self, epoch, step):
        self.net.eval()  # train mode
        loss_all = 0
        with torch.no_grad():
            with tqdm(total=self.data.shape[0], desc=f'[DRE] Validating.. Epoch {epoch + 1}/{self.num_epochs}', unit='img',
                      colour='green') as pbar:
                for i in range(self.num_batch):
                    # Load the packed data:
                    tar = self.P[i]
                    inputs, targets = torch.from_numpy(self.data[i*self.batch_size:(i+1)*self.batch_size]).type(torch.FloatTensor).to(self.device), \
                                      torch.from_numpy(tar).type(torch.FloatTensor).to(self.device)
                    inputs = inputs.reshape(inputs.shape[0], -1)
                    _, _, _, outputs = self.net(inputs)
                    # loss = kl_divergence_bayes(outputs, targets, 1, knn_bayes)
                    loss = loss_function(targets, outputs, self.a, self.b, type=step)
                    loss_all += loss
                    # loss_aver = loss_all / (i+1)
                    pbar.set_postfix(**{'loss (batch)': loss_all / self.num_batch})
                    pbar.update(targets.shape[0])
        # Save checkpoint.
        if float(loss_all / self.num_batch) < self.lowest_loss:
            print('[DRE] Best accuracy, saving the weights...')
            self.net_optim = self.net
            lowest_loss = float(loss_all / self.num_batch)
        return loss_all / self.num_batch

    def _validation_conv(self, epoch, step):
        self.net.eval()  # train mode
        loss_all = 0
        with torch.no_grad():
            with tqdm(total=self.data.shape[0], desc=f'[DRE] Validating.. Epoch {epoch + 1}/{self.num_epochs}', unit='img',
                      colour='green') as pbar:
                for i in range(self.num_batch):
                    # Load the packed data:
                    tar = self.P[i]
                    inputs, targets = torch.from_numpy(self.data[i*self.batch_size:(i+1)*self.batch_size])\
                                          .type(torch.FloatTensor).permute(0, 3, 1, 2).to(self.device), \
                                      torch.from_numpy(tar).type(torch.FloatTensor).to(self.device)  # 如果是long会全变成0
                    _, _, _, outputs = self.net(inputs)
                    # loss = kl_divergence_bayes(outputs, targets, 1, knn_bayes)
                    loss = loss_function(targets, outputs, self.a, self.b, type=step)
                    loss_all += loss
                    # loss_aver = loss_all / (i+1)
                    pbar.set_postfix(**{'loss (batch)': loss_all / self.num_batch})
                    pbar.update(targets.shape[0])
        # Save checkpoint.
        if float(loss_all / self.num_batch) < self.lowest_loss:
            print('[DRE] Best accuracy, saving the weights...')
            self.net_optim = self.net
            lowest_loss = float(loss_all / self.num_batch)
        return loss_all / self.num_batch

    def plot(self, epoch, step):
        self.embedding = self._plot(epoch, step)

    def _plot(self, epoch, step, fig_size='normal'):
        self.net.eval()
        with torch.no_grad():
            for i in range(self.num_batch):
                if self.dre_type == 'conv':
                    inputs = torch.from_numpy(self.data[i * self.batch_size:(i + 1) * self.batch_size]).type(
                                                    torch.FloatTensor)  # conv
                elif self.dre_type == 'fc':
                    inputs = torch.from_numpy(self.data[i * self.batch_size:(i + 1) * self.batch_size])\
                        .type(torch.FloatTensor)
                else:
                    raise TypeError('DRE type must be conv or fc.')

                if i == 0:
                    _, _, _, Y = self.net(inputs)
                    Y = Y.cpu()
                else:
                    _, _, _, y_test = self.net(inputs)
                    Y = np.concatenate((Y, y_test.cpu()), axis=0)
        torch.cuda.empty_cache()
        plt.figure(figsize=(10, 10))
        if fig_size == 'fixed':
            plt.xlim([-500, 500])
            plt.ylim([-500, 500])
        scatter = plt.scatter(Y[:, 0], Y[:, 1], s=1, c='darkorange', alpha=0.7)
        # legend1 = plt.legend(*scatter.legend_elements(), title="Classes", loc='upper right')
        plt.title('Epoch = %d, Loss = %f' % (epoch, self.loss_score_train))
        plt.tight_layout()
        # plt.scatter(Y[:, 0], Y[:, 1], s=1, c=targets)
        # plt.scatter(Y[:, 0], Y[:, 1], s=1, c=plt.cm.cubehelix(0.1 * targets))
        # plt.savefig("./DRE_labeled_%s_%s.png" % (time.asctime(time.localtime(time.time())), step))
        if self.plot_results:
            plt.savefig(self.directory + "DRE_labeled_%s_%s.png" % (time.asctime(time.localtime(time.time())), step))
        if step == 're_umap':
            return Y
        else:
            return 0

    def _fit_embedding(self):
        start_time = time.time()
        print('[DRE] start------->  time: ', time.ctime(time.time()))
        recursive_step = 'pre'
        for epoch in range(self.num_epochs):

            if epoch == 0:
                self.P = self.calculate_p_matrix(recursive_step)

            if epoch % self.rolling_num == 0 and epoch != 0 and epoch not in \
                    np.int16(self.num_pre_epochs + np.int16([0, 1, 2, 3])*self.num_recursive_epochs):
                self.P = self.calculate_p_matrix(recursive_step)

            if epoch == self.num_pre_epochs:  # 300
                if self.save_steps:
                # save the model:
                    state = {
                        'net': self.net.state_dict(),
                        'loss': self.loss_score_train,
                        'epoch': epoch,
                    }
                    if not os.path.isdir(self.directory+'DRE_model_checkpoint'):
                        os.mkdir(self.directory+'DRE_model_checkpoint')
                    torch.save(state, self.directory+'DRE_model_checkpoint/DRE_{}.pth'.format(recursive_step))
                if self.plot_results:
                    self.plot(epoch, recursive_step)

                recursive_step = 're1'
                # calculate the new P matrix:
                self.P = self.calculate_p_matrix(recursive_step)

            if epoch == self.num_pre_epochs + self.num_recursive_epochs:  # 400
                if self.save_steps:
                # save the model:
                    state = {
                        'net': self.net.state_dict(),
                        'loss': self.loss_score_train,
                        'epoch': epoch,
                    }
                    torch.save(state, self.directory+'DRE_model_checkpoint/DRE_{}.pth'.format(recursive_step))

                if self.plot_results:
                    self.plot(epoch, recursive_step)

                recursive_step = 're2'
                # calculate the new P matrix:
                self.P = self.calculate_p_matrix(recursive_step)

            if epoch == self.num_pre_epochs + 2*self.num_recursive_epochs:  # 500
                if self.save_steps:
                    # save the model:
                    state = {
                        'net': self.net.state_dict(),
                        'loss': self.loss_score_train,
                        'epoch': epoch,
                    }
                    torch.save(state, self.directory+'DRE_model_checkpoint/DRE_{}.pth'.format(recursive_step))
                
                if self.plot_results:
                    self.plot(epoch, recursive_step)

                recursive_step = 're3'
                # calculate the new P matrix:
                self.P = self.calculate_p_matrix(recursive_step)

            if epoch == self.num_pre_epochs + 3*self.num_recursive_epochs:  # 600
                if self.save_steps:
                    # save the model:
                    state = {
                        'net': self.net.state_dict(),
                        'loss': self.loss_score_train,
                        'epoch': epoch,
                    }
                    torch.save(state, self.directory+'DRE_model_checkpoint/DRE_{}.pth'.format(recursive_step))

                if self.plot_results:
                    self.plot(epoch, recursive_step)

                recursive_step = 're_umap'
                # calculate the new P matrix:
                self.P = self.calculate_p_matrix(recursive_step)

            # <==================== train ====================>

            if self.dre_type == 'conv':
                self.loss_score_train = self._train_conv(epoch, recursive_step)
            elif self.dre_type == 'fc':
                self.loss_score_train = self._train(epoch, recursive_step)
            else:
                raise TypeError('DRE type should be conv or fc.')

            # <==================== validate ====================>

            if (epoch + 1) % 10 == 0:
                if self.dre_type == 'conv':
                    self.loss_score_val = self._validation_conv(epoch, recursive_step).cpu()
                elif self.dre_type == 'fc':
                    self.loss_score_val = self._validation(epoch, recursive_step).cpu()
                else:
                    raise TypeError('DRE type should be conv or fc.')
                self.loss_plot_val.append(self.loss_score_val)
            # scheduler.step()
            self.loss_plot_train.append(self.loss_score_train)

        end_time = time.time()
        duration = end_time - start_time
        print('[DRE] training time: ', duration)
        # state = {
        #     'net': self.net.state_dict(),
        #     'loss': self.loss_score_train,
        #     'epoch': epoch,
        # }
        # torch.save(state, './checkpoint/DRE_{}.pth'.format(recursive_step))

        print('[DRE] ------->complete.  time: ', time.ctime(time.time()))

        # # plot validation loss
        # plt.clf()
        # # plt.style.use(style='Solarize_Light2')
        # plt.plot(self.loss_plot_val, color='b')
        # plt.title('Loss in the iteration (validation). Epochs = %d, Learning rate: %f' % (self.num_epochs, self.lr))
        # plt.xlabel('epochs / 10')
        # plt.ylabel('loss')
        # plt.savefig("./loss_val_{}.png".format(recursive_step))

        # # plot training loss
        # plt.clf()
        # plt.plot(self.loss_plot_train, color='b')
        # plt.title('Loss in the iteration (training). Epochs = %d, Learning rate: %f' % (self.num_epochs, self.lr))
        # plt.xlabel('epochs')
        # plt.ylabel('loss')
        # plt.savefig("./loss_train_{}.png".format(recursive_step))

        self.plot(epoch, recursive_step)

    def fit_transform(self, x):
        start_time = time.time()
        self.data = x
        self.num_batch = ceil(self.data.shape[0] / self.batch_size)
        self._fit_embedding()
        end_time = time.time()
        print('fitting time: {}s'.format(end_time - start_time))

        return self.embedding

    def save_model(self):
        state = {
            'net': self.net_optim.state_dict(),
            'loss': loss.item(),
            'epoch': epoch,
        }
        if not os.path.isdir(self.directory+'DRE_model_checkpoint'):
            os.mkdir(self.directory+'DRE_model_checkpoint')
        # Remember to rename the saved model corresponding to the hyper-parameters:
        torch.save(state, self.directory+'DRE_model_checkpoint/DRE_manually_save.pth')






import os

import numpy


def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration("dre", parent_package, top_path)

    return config


if __name__ == "__main__":
    from numpy.distutils.core import setup

    setup(**configuration().todict())


# Author: Leland McInnes <leland.mcinnes@gmail.com>
#
# License: BSD 3 clause
from __future__ import print_function

import locale
from warnings import warn
import time

from scipy.optimize import curve_fit

try:
    import joblib
except ImportError:
    # sklearn.externals.joblib is deprecated in 0.21, will be removed in 0.23
    from sklearn.externals import joblib

import numpy as np
import scipy.sparse
from scipy.sparse import tril as sparse_tril, triu as sparse_triu
import scipy.sparse.csgraph
import numba

from umap_xinrui.utils import (
    submatrix,
    ts,
    csr_unique,
    fast_knn_indices,
)

from pynndescent import NNDescent

locale.setlocale(locale.LC_NUMERIC, "C")

INT32_MIN = np.iinfo(np.int32).min + 1
INT32_MAX = np.iinfo(np.int32).max - 1

SMOOTH_K_TOLERANCE = 1e-5  # 1e-5
MIN_K_DIST_SCALE = 1e-3
NPY_INFINITY = np.inf
MACHINE_EPSILON = np.finfo(np.double).eps

DISCONNECTION_DISTANCES = {
    "correlation": 1,
    "cosine": 1,
    "hellinger": 1,
    "jaccard": 1,
    "dice": 1,
}


def flatten_iter(container):
    for i in container:
        if isinstance(i, (list, tuple)):
            for j in flatten_iter(i):
                yield j
        else:
            yield i


def flattened(container):
    return tuple(flatten_iter(container))


def breadth_first_search(adjmat, start, min_vertices):
    explored = []
    queue = [start]
    levels = {}
    levels[start] = 0
    max_level = np.inf
    visited = [start]

    while queue:
        node = queue.pop(0)
        explored.append(node)
        if max_level == np.inf and len(explored) > min_vertices:
            max_level = max(levels.values())

        if levels[node] + 1 < max_level:
            neighbors = adjmat[node].indices
            for neighbour in neighbors:
                if neighbour not in visited:
                    queue.append(neighbour)
                    visited.append(neighbour)

                    levels[neighbour] = levels[node] + 1

    return np.array(explored)


def raise_disconnected_warning(
    edges_removed,
    vertices_disconnected,
    disconnection_distance,
    total_rows,
    threshold=0.1,
    verbose=False,
):
    """A simple wrapper function to avoid large amounts of code repetition."""
    if verbose & (vertices_disconnected == 0) & (edges_removed > 0):
        print(
            f"Disconnection_distance = {disconnection_distance} has removed {edges_removed} edges.  "
            f"This is not a problem as no vertices were disconnected."
        )
    elif (vertices_disconnected > 0) & (
        vertices_disconnected <= threshold * total_rows
    ):
        warn(
            f"A few of your vertices were disconnected from the manifold.  This shouldn't cause problems.\n"
            f"Disconnection_distance = {disconnection_distance} has removed {edges_removed} edges.\n"
            f"It has only fully disconnected {vertices_disconnected} vertices.\n"
            f"Use umap.utils.disconnected_vertices() to identify them.",
        )
    elif vertices_disconnected > threshold * total_rows:
        warn(
            f"A large number of your vertices were disconnected from the manifold.\n"
            f"Disconnection_distance = {disconnection_distance} has removed {edges_removed} edges.\n"
            f"It has fully disconnected {vertices_disconnected} vertices.\n"
            f"You might consider using find_disconnected_points() to find and remove these points from your data.\n"
            f"Use umap.utils.disconnected_vertices() to identify them.",
        )


@numba.njit(
    locals={
        "psum": numba.types.float32,
        "lo": numba.types.float32,
        "mid": numba.types.float32,
        "hi": numba.types.float32,
    },
    fastmath=True,
)  # benchmarking `parallel=True` shows it to *decrease* performance
def smooth_knn_dist(distances, k, n_iter=64, local_connectivity=1.0, bandwidth=1.0):  # n_iter=64
    """Compute a continuous version of the distance to the kth nearest
    neighbor. That is, this is similar to knn-distance but allows continuous
    k values rather than requiring an integral k. In essence we are simply
    computing the distance such that the cardinality of fuzzy set we generate
    is k.

    Parameters
    ----------
    distances: array of shape (n_samples, n_neighbors)
        Distances to nearest neighbors for each samples. Each row should be a
        sorted list of distances to a given samples nearest neighbors.

    k: float
        The number of nearest neighbors to approximate for.

    n_iter: int (optional, default 64)
        We need to binary search for the correct distance value. This is the
        max number of iterations to use in such a search.

    local_connectivity: int (optional, default 1)
        The local connectivity required -- i.e. the number of nearest
        neighbors that should be assumed to be connected at a local level.
        The higher this value the more connected the manifold becomes
        locally. In practice this should be not more than the local intrinsic
        dimension of the manifold.

    bandwidth: float (optional, default 1)
        The target bandwidth of the kernel, larger values will produce
        larger return values.

    Returns
    -------
    knn_dist: array of shape (n_samples,)
        The distance to kth nearest neighbor, as suitably approximated.

    nn_dist: array of shape (n_samples,)
        The distance to the 1st nearest neighbor for each point.
    """
    target = np.log2(k) * bandwidth
    rho = np.zeros(distances.shape[0], dtype=np.float32)
    result = np.zeros(distances.shape[0], dtype=np.float32)

    mean_distances = np.mean(distances)

    for i in range(distances.shape[0]):
        lo = 0.0
        hi = NPY_INFINITY
        mid = 1.0

        # TODO: This is very inefficient, but will do for now. FIXME
        ith_distances = distances[i]
        non_zero_dists = ith_distances[ith_distances > 0.0]
        if non_zero_dists.shape[0] >= local_connectivity:
            index = int(np.floor(local_connectivity))
            interpolation = local_connectivity - index
            if index > 0:
                rho[i] = non_zero_dists[index - 1]
                if interpolation > SMOOTH_K_TOLERANCE:
                    rho[i] += interpolation * (
                        non_zero_dists[index] - non_zero_dists[index - 1]
                    )
            else:
                rho[i] = interpolation * non_zero_dists[0]
                # rho[i] = interpolation * non_zero_dists[1]  # for bayes!
        elif non_zero_dists.shape[0] > 0:
            rho[i] = np.max(non_zero_dists)

        for n in range(n_iter):

            psum = 0.0
            for j in range(1, distances.shape[1]):
                d = distances[i, j] - rho[i]
                if d > 0:
                    psum += np.exp(-(d / mid))  # TODO: 这里似乎已经计算了probability，可以再优化一下
                else:
                    psum += 1.0

            if np.fabs(psum - target) < SMOOTH_K_TOLERANCE:
                break

            if psum > target:
                hi = mid
                mid = (lo + hi) / 2.0
            else:
                lo = mid
                if hi == NPY_INFINITY:
                    mid *= 2
                else:
                    mid = (lo + hi) / 2.0

        result[i] = mid

        # TODO: This is very inefficient, but will do for now. FIXME
        if rho[i] > 0.0:
            mean_ith_distances = np.mean(ith_distances)
            if result[i] < MIN_K_DIST_SCALE * mean_ith_distances:
                result[i] = MIN_K_DIST_SCALE * mean_ith_distances
        else:
            if result[i] < MIN_K_DIST_SCALE * mean_distances:
                result[i] = MIN_K_DIST_SCALE * mean_distances

    return result, rho


@numba.njit(
    locals={
        "psum": numba.types.float32,
        "lo": numba.types.float32,
        "mid": numba.types.float32,
        "hi": numba.types.float32,
        "Pj": numba.types.float32[::1],
        "distances_bayes": numba.types.float32[::1],
    },
    fastmath=True,
)  # benchmarking `parallel=True` shows it to *decrease* performance
def smooth_knn_dist_v2(distances, knn_indices, k, n_iter=64, local_connectivity=1.0, bandwidth=1.0):
    target = np.log2(k) * bandwidth
    rho = np.zeros(distances.shape[0], dtype=np.float32)
    result = np.zeros(distances.shape[0], dtype=np.float32)

    for i in range(distances.shape[0]):
        lo = 0.0
        hi = NPY_INFINITY
        mid = 1.0

        # TODO: This is very inefficient, but will do for now. FIXME
        # Find rhos:
        rho[i] = np.min(distances[1:])

        for n in range(n_iter):

            psum = 0.0
            for j in range(1, distances.shape[1]):
                d = distances[i, j] - rho[i]  # rho到底是在bayes之前还是之后？
                if d > 0:
                    psum += np.exp(-(d / mid))  # TODO: 这里似乎已经计算了probability，可以再优化一下
                else:
                    psum += 1.0

            if np.fabs(psum - target) < SMOOTH_K_TOLERANCE:
                break

            if psum > target:
                hi = mid
                mid = (lo + hi) / 2.0
            else:
                lo = mid
                if hi == NPY_INFINITY:
                    mid *= 2
                else:
                    mid = (lo + hi) / 2.0

        result[i] = mid

        # # TODO: This is very inefficient, but will do for now. FIXME
        # if rho[i] > 0.0:
        #     mean_ith_distances = np.mean(ith_distances)
        #     if result[i] < MIN_K_DIST_SCALE * mean_ith_distances:
        #         result[i] = MIN_K_DIST_SCALE * mean_ith_distances
        # else:
        #     if result[i] < MIN_K_DIST_SCALE * mean_distances:
        #         result[i] = MIN_K_DIST_SCALE * mean_distances

    return result, rho


@numba.njit(
    locals={
        "psum": numba.types.float32,
        "lo": numba.types.float32,
        "mid": numba.types.float32,
        "hi": numba.types.float32,
    },
    fastmath=True,
)  # benchmarking `parallel=True` shows it to *decrease* performance
def smooth_knn_dist_t_dist(distances, k, n_iter=64, local_connectivity=1.0, bandwidth=1.0):
    target = np.log2(k) * bandwidth
    rho = np.zeros(distances.shape[0], dtype=np.float32)
    result = np.zeros(distances.shape[0], dtype=np.float32)

    mean_distances = np.mean(distances)

    for i in range(distances.shape[0]):
        lo = 0.0
        hi = NPY_INFINITY
        mid = 1.0

        # TODO: This is very inefficient, but will do for now. FIXME
        ith_distances = distances[i]
        non_zero_dists = ith_distances[ith_distances > 0.0]
        if non_zero_dists.shape[0] >= local_connectivity:
            index = int(np.floor(local_connectivity))
            interpolation = local_connectivity - index
            if index > 0:
                rho[i] = non_zero_dists[index - 1]
                if interpolation > SMOOTH_K_TOLERANCE:
                    rho[i] += interpolation * (
                        non_zero_dists[index] - non_zero_dists[index - 1]
                    )
            else:
                rho[i] = interpolation * non_zero_dists[0]
        elif non_zero_dists.shape[0] > 0:
            rho[i] = np.max(non_zero_dists)

        for n in range(n_iter):

            psum = 0.0
            for j in range(1, distances.shape[1]):
                d = distances[i, j] - rho[i]
                if d > 0:
                    psum += np.power((d / mid), -1)  # TODO: 这里似乎已经计算了probability，可以再优化一下
                else:
                    psum += 1.0

            if np.fabs(psum - target) < SMOOTH_K_TOLERANCE:
                break

            if psum > target:
                hi = mid
                mid = (lo + hi) / 2.0
            else:
                lo = mid
                if hi == NPY_INFINITY:
                    mid *= 2
                else:
                    mid = (lo + hi) / 2.0

        result[i] = mid

        # TODO: This is very inefficient, but will do for now. FIXME
        if rho[i] > 0.0:
            mean_ith_distances = np.mean(ith_distances)
            if result[i] < MIN_K_DIST_SCALE * mean_ith_distances:
                result[i] = MIN_K_DIST_SCALE * mean_ith_distances
        else:
            if result[i] < MIN_K_DIST_SCALE * mean_distances:
                result[i] = MIN_K_DIST_SCALE * mean_distances

    return result, rho


@numba.njit(
    locals={
        "psum": numba.types.float32,
        "lo": numba.types.float32,
        "mid": numba.types.float32,
        "hi": numba.types.float32,
    },
    fastmath=True,
)  # benchmarking `parallel=True` shows it to *decrease* performance
def smooth_knn_dist_bayes(distances, k, n_iter=64, local_connectivity=1.0, bandwidth=1.0):  # n_iter=64
    target = np.log2(k) * bandwidth
    rho = np.zeros(distances.shape[0], dtype=np.float32)
    result = np.zeros(distances.shape[0], dtype=np.float32)

    mean_distances = np.mean(distances)

    for i in range(distances.shape[0]):
        lo = 0.0
        hi = NPY_INFINITY
        mid = 1.0

        # TODO: This is very inefficient, but will do for now. FIXME
        ith_distances = distances[i]
        non_zero_dists = ith_distances[ith_distances > 0.0]
        if non_zero_dists.shape[0] >= local_connectivity:
            index = int(np.floor(local_connectivity))
            interpolation = local_connectivity - index
            if index > 0:
                rho[i] = non_zero_dists[index - 1]
                if interpolation > SMOOTH_K_TOLERANCE:
                    rho[i] += interpolation * (
                        non_zero_dists[index] - non_zero_dists[index - 1]
                    )
            else:
                rho[i] = interpolation * non_zero_dists[1]
                # rho[i] = interpolation * non_zero_dists[1]  # for bayes!
        elif non_zero_dists.shape[0] > 0:
            rho[i] = np.max(non_zero_dists)

        for n in range(n_iter):

            psum = 0.0
            for j in range(1, distances.shape[1]):
                d = distances[i, j] - rho[i]
                if d > 0:
                    psum += np.exp(-(d / mid))  # TODO: 这里似乎已经计算了probability，可以再优化一下
                else:
                    psum += 1.0

            if np.fabs(psum - target) < SMOOTH_K_TOLERANCE:
                break

            if psum > target:
                hi = mid
                mid = (lo + hi) / 2.0
            else:
                lo = mid
                if hi == NPY_INFINITY:
                    mid *= 2
                else:
                    mid = (lo + hi) / 2.0

        result[i] = mid

        # TODO: This is very inefficient, but will do for now. FIXME
        if rho[i] > 0.0:
            mean_ith_distances = np.mean(ith_distances)
            if result[i] < MIN_K_DIST_SCALE * mean_ith_distances:
                result[i] = MIN_K_DIST_SCALE * mean_ith_distances
        else:
            if result[i] < MIN_K_DIST_SCALE * mean_distances:
                result[i] = MIN_K_DIST_SCALE * mean_distances

    return result, rho


def nearest_neighbors(
    X,
    n_neighbors,
    metric,
    metric_kwds,
    angular,
    random_state,
    low_memory=True,
    use_pynndescent=True,
    n_jobs=-1,
    verbose=False,
):
    if verbose:
        print(ts(), "Finding Nearest Neighbors")

    if metric == "precomputed":
        # Note that this does not support sparse distance matrices yet ...
        # Compute indices of n nearest neighbors
        knn_indices = fast_knn_indices(X, n_neighbors)
        # knn_indices = np.argsort(X)[:, :n_neighbors]
        # Compute the nearest neighbor distances
        #   (equivalent to np.sort(X)[:,:n_neighbors])
        knn_dists = X[np.arange(X.shape[0])[:, None], knn_indices].copy()
        # Prune any nearest neighbours that are infinite distance apart.
        disconnected_index = knn_dists == np.inf
        knn_indices[disconnected_index] = -1

        knn_search_index = None
    else:
        # TODO: Hacked values for now
        n_trees = min(64, 5 + int(round((X.shape[0]) ** 0.5 / 20.0)))
        n_iters = max(5, int(round(np.log2(X.shape[0]))))

        knn_search_index = NNDescent(
            X,
            n_neighbors=n_neighbors,
            metric=metric,
            metric_kwds=metric_kwds,
            random_state=random_state,
            n_trees=n_trees,
            n_iters=n_iters,
            max_candidates=60,
            low_memory=low_memory,
            n_jobs=n_jobs,
            verbose=verbose,
        )
        knn_indices, knn_dists = knn_search_index.neighbor_graph

    if verbose:
        print(ts(), "Finished Nearest Neighbor Search")
    return knn_indices, knn_dists, knn_search_index


@numba.njit(
    locals={
        "knn_dists": numba.types.float32[:, ::1],
        "sigmas": numba.types.float32[::1],
        "rhos": numba.types.float32[::1],
        "val": numba.types.float32,
    },
    parallel=True,
    fastmath=True,
)
def compute_membership_strengths(
    knn_indices, knn_dists, sigmas, rhos, return_dists=False, bipartite=False,
):
    n_samples = knn_indices.shape[0]
    n_neighbors = knn_indices.shape[1]

    rows = np.zeros(knn_indices.size, dtype=np.int32)
    cols = np.zeros(knn_indices.size, dtype=np.int32)
    vals = np.zeros(knn_indices.size, dtype=np.float32)
    if return_dists:
        dists = np.zeros(knn_indices.size, dtype=np.float32)
    else:
        dists = None

    for i in range(n_samples):
        for j in range(n_neighbors):
            if knn_indices[i, j] == -1:
                continue  # We didn't get the full knn for i
            # If applied to an adjacency matrix points shouldn't be similar to themselves.
            # If applied to an incidence matrix (or bipartite) then the row and column indices are different.
            if (bipartite == False) & (knn_indices[i, j] == i):
                val = 0.0
            elif knn_dists[i, j] - rhos[i] <= 0.0 or sigmas[i] == 0.0:
                val = 1.0
            else:
                # *******************************************************************************
                # ******************************** main function ********************************
                # *******************************************************************************
                val = np.exp(-((knn_dists[i, j] - rhos[i]) / (sigmas[i])))

            rows[i * n_neighbors + j] = i
            cols[i * n_neighbors + j] = knn_indices[i, j]
            vals[i * n_neighbors + j] = val
            if return_dists:
                dists[i * n_neighbors + j] = knn_dists[i, j]

    return rows, cols, vals, dists


@numba.njit(
    locals={
        "knn_dists": numba.types.float32[:, ::1],
        "sigmas": numba.types.float32[::1],
        "rhos": numba.types.float32[::1],
        "val": numba.types.float32,
    },
    parallel=True,
    fastmath=True,
)
def compute_membership_strengths_t_dist(
    knn_indices, knn_dists, sigmas, rhos, return_dists=False, bipartite=False,
):
    n_samples = knn_indices.shape[0]
    n_neighbors = knn_indices.shape[1]

    rows = np.zeros(knn_indices.size, dtype=np.int32)
    cols = np.zeros(knn_indices.size, dtype=np.int32)
    vals = np.zeros(knn_indices.size, dtype=np.float32)
    if return_dists:
        dists = np.zeros(knn_indices.size, dtype=np.float32)
    else:
        dists = None

    for i in range(n_samples):
        for j in range(n_neighbors):
            if knn_indices[i, j] == -1:
                continue  # We didn't get the full knn for i
            # If applied to an adjacency matrix points shouldn't be similar to themselves.
            # If applied to an incidence matrix (or bipartite) then the row and column indices are different.
            if (bipartite == False) & (knn_indices[i, j] == i):
                val = 0.0
            elif knn_dists[i, j] - rhos[i] <= 0.0 or sigmas[i] == 0.0:
                val = 1.0
            else:

                # *******************************************************************************
                # ******************************** main function ********************************
                # *******************************************************************************

                val = np.power((knn_dists[i, j] - rhos[i]) / (sigmas[i]), 1.0 + 1.0 / -2.0)

            rows[i * n_neighbors + j] = i
            cols[i * n_neighbors + j] = knn_indices[i, j]
            vals[i * n_neighbors + j] = val
            if return_dists:
                dists[i * n_neighbors + j] = knn_dists[i, j]

    return rows, cols, vals, dists


def fuzzy_simplicial_set(
    X,
    n_neighbors,
    random_state,
    metric,
    metric_kwds={},
    knn_indices=None,
    knn_dists=None,
    angular=False,
    set_op_mix_ratio=1.0,
    local_connectivity=1.0,
    apply_set_operations=True,
    verbose=False,
    return_dists=None,
):
    if knn_indices is None or knn_dists is None:

        # numba:
        knn_indices, knn_dists, _ = nearest_neighbors(
            X, n_neighbors, metric, metric_kwds, angular, random_state, verbose=verbose,
        )

    knn_dists = knn_dists.astype(np.float32)

    # numba: This step already compute the P matrix inplicitly, TODO: extract this P matrix
    # only use knn_dists
    sigmas, rhos = smooth_knn_dist(
        knn_dists, float(n_neighbors), local_connectivity=float(local_connectivity),
    )

    # numba:
    # use knn_dists and knn_indices
    rows, cols, vals, dists = compute_membership_strengths(
        knn_indices, knn_dists, sigmas, rhos, return_dists
    )

    result = scipy.sparse.coo_matrix(
        (vals, (rows, cols)), shape=(X.shape[0], X.shape[0])
    )
    result.eliminate_zeros()

    if apply_set_operations:
        transpose = result.transpose()

        prod_matrix = result.multiply(transpose)

        result = (
            set_op_mix_ratio * (result + transpose - prod_matrix)
            + (1.0 - set_op_mix_ratio) * prod_matrix
        )

    result.eliminate_zeros()

    if return_dists is None:
        return result, sigmas, rhos
    else:
        if return_dists:
            dmat = scipy.sparse.coo_matrix(
                (dists, (rows, cols)), shape=(X.shape[0], X.shape[0])
            )

            dists = dmat.maximum(dmat.transpose()).todok()
        else:
            dists = None

        return result, sigmas, rhos, dists


def make_epochs_per_sample(weights, n_epochs):
    """Given a set of weights and number of epochs generate the number of
    epochs per sample for each weight.

    Parameters
    ----------
    weights: array of shape (n_1_simplices)
        The weights ofhow much we wish to sample each 1-simplex.

    n_epochs: int
        The total number of epochs we want to train for.

    Returns
    -------
    An array of number of epochs per sample, one for each 1-simplex.
    """
    result = -1.0 * np.ones(weights.shape[0], dtype=np.float64)
    n_samples = n_epochs * (weights / weights.max())
    result[n_samples > 0] = float(n_epochs) / n_samples[n_samples > 0]
    return result


@numba.njit()
def init_transform(indices, weights, embedding):
    """Given indices and weights and an original embeddings
    initialize the positions of new points relative to the
    indices and weights (of their neighbors in the source data).

    Parameters
    ----------
    indices: array of shape (n_new_samples, n_neighbors)
        The indices of the neighbors of each new sample

    weights: array of shape (n_new_samples, n_neighbors)
        The membership strengths of associated 1-simplices
        for each of the new samples.

    embedding: array of shape (n_samples, dim)
        The original embedding of the source data.

    Returns
    -------
    new_embedding: array of shape (n_new_samples, dim)
        An initial embedding of the new sample points.
    """
    result = np.zeros((indices.shape[0], embedding.shape[1]), dtype=np.float32)

    for i in range(indices.shape[0]):
        for j in range(indices.shape[1]):
            for d in range(embedding.shape[1]):
                result[i, d] += weights[i, j] * embedding[indices[i, j], d]

    return result


def init_graph_transform(graph, embedding):
    """Given a bipartite graph representing the 1-simplices and strengths between the
     new points and the original data set along with an embedding of the original points
    initialize the positions of new points relative to the strengths (of their neighbors in the source data).

    If a point is in our original data set it embeds at the original points coordinates.
    If a point has no neighbours in our original dataset it embeds as the np.nan vector.
    Otherwise a point is the weighted average of it's neighbours embedding locations.

    Parameters
    ----------
    graph: csr_matrix (n_new_samples, n_samples)
        A matrix indicating the the 1-simplices and their associated strengths.  These strengths should
        be values between zero and one and not normalized.  One indicating that the new point was identical
        to one of our original points.

    embedding: array of shape (n_samples, dim)
        The original embedding of the source data.

    Returns
    -------
    new_embedding: array of shape (n_new_samples, dim)
        An initial embedding of the new sample points.
    """
    result = np.zeros((graph.shape[0], embedding.shape[1]), dtype=np.float32)

    for row_index in range(graph.shape[0]):
        num_neighbours = len(graph[row_index].indices)
        if num_neighbours == 0:
            result[row_index] = np.nan
            continue
        for col_index in graph[row_index].indices:
            if graph[row_index, col_index] == 1:
                result[row_index, :] = embedding[col_index, :]
                break
            for d in range(embedding.shape[1]):
                result[row_index, d] += (
                    graph[row_index, col_index]
                    / num_neighbours
                    * embedding[col_index, d]
                )

    return result


@numba.njit()
def init_update(current_init, n_original_samples, indices):
    for i in range(n_original_samples, indices.shape[0]):
        n = 0
        for j in range(indices.shape[1]):
            for d in range(current_init.shape[1]):
                if indices[i, j] < n_original_samples:
                    n += 1
                    current_init[i, d] += current_init[indices[i, j], d]
        for d in range(current_init.shape[1]):
            current_init[i, d] /= n

    return


def find_ab_params(spread, min_dist):
    """Fit a, b params for the differentiable curve used in lower
    dimensional fuzzy simplicial complex construction. We want the
    smooth curve (from a pre-defined family with simple gradient) that
    best matches an offset exponential decay.
    """

    def curve(x, a, b):
        return 1.0 / (1.0 + a * x ** (2 * b))

    xv = np.linspace(0, spread * 3, 300)
    yv = np.zeros(xv.shape)
    yv[xv < min_dist] = 1.0
    yv[xv >= min_dist] = np.exp(-(xv[xv >= min_dist] - min_dist) / spread)
    params, covar = curve_fit(curve, xv, yv)
    return params[0], params[1]



import os
import numpy
from setuptools import find_packages, setup
from numpy.distutils.core import Extension
from Cython.Build import cythonize


def readme():
    with open('README.md') as readme_file:
        return readme_file.read()

os.environ["CC"] = "clang"

configuration = {
    'name' : 'DRE',
    'version': '1.0.9',
    'description' : 'Deep Recursive Embedding for High-Dimensional Data',
    'long_description' : readme(),
    'classifiers' : [
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        "License :: OSI Approved :: MIT License",
    ],
    'keywords' : 'Dimensionality Reduction t-SNE UMAP Representation Learning',
    'url' : 'https://github.com/zuxinrui/DRE',
    'author' : 'Xinrui Zu',
    'author_email' : 'zuxinrui95@gmail.com',
    'license' : 'LICENSE',
    'packages' : find_packages(),
    'install_requires' : ['scikit-learn >= 0.16',
                          'numba >= 0.34',
                          'torch >= 1.0',],
    'ext_modules' : cythonize(['DRE/_utils_tsne.pyx'],),
    'include_dirs' : [numpy.get_include()],
    }

setup(**configuration)
